@@ -53,6 +53,13 @@
 /* We assume here int holds at least 32 bits */
 static int *red_table = DUMMY, *green_table = DUMMY, *blue_table = DUMMY;
 static int table_16 = 1;
+static unsigned short *
+#ifdef __GNUC__
+#if __GNUC__ == 3
+volatile
+#endif
+#endif
+real_colors_table = NULL;
 
 /* If we want to represent some 16-bit from-screen-light, it would require certain display input
  * value (0-255 red, 0-255 green, 0-255 blue), possibly not a whole number. [red|green|blue]_table
@@ -94,66 +101,25 @@ void (*round_fn)(unsigned short *my_restrict in, struct bitmap *out);
 /* When you finish the stuff with dither_start, dither_restart, just do "if (dregs) mem_free(dregs);" */
 static void (*dither_fn_internal)(unsigned short *my_restrict in, struct bitmap *out, int *dregs);
 
-
-     /* prototypes */
-static void dither_1byte(unsigned short *my_restrict, struct bitmap *, int *);   /* DITHER_TEMPLATE */
-static void round_1byte(unsigned short *my_restrict, struct bitmap *);           /* ROUND_TEMPLATE */
-static void dither_2byte(unsigned short *my_restrict, struct bitmap *, int *);   /* DITHER_TEMPLATE */
-static void round_2byte(unsigned short *my_restrict, struct bitmap *);           /* ROUND_TEMPLATE */
-static void dither_195(unsigned short *my_restrict, struct bitmap *, int *);   /* DITHER_TEMPLATE */
-static void round_195(unsigned short *my_restrict, struct bitmap *);           /* ROUND_TEMPLATE */
-static void dither_451(unsigned short *my_restrict, struct bitmap *, int *);   /* DITHER_TEMPLATE */
-static void round_451(unsigned short *my_restrict, struct bitmap *);           /* ROUND_TEMPLATE */
-static void dither_196(unsigned short *my_restrict, struct bitmap *, int *);   /* DITHER_TEMPLATE */
-static void round_196(unsigned short *my_restrict, struct bitmap *);           /* ROUND_TEMPLATE */
-static void dither_452(unsigned short *my_restrict, struct bitmap *, int *);   /* DITHER_TEMPLATE */
-static void round_452(unsigned short *my_restrict, struct bitmap *);           /* ROUND_TEMPLATE */
-static void dither_708(unsigned short *my_restrict, struct bitmap *, int *);   /* DITHER_TEMPLATE */
-static void round_708(unsigned short *my_restrict, struct bitmap *);           /* ROUND_TEMPLATE */
-static long color_332(int);
-static long color_121(int);
-static long color_pass_rgb(int);
-static long color_888_bgr(int);
-/*static void pass_bgr(unsigned short *, struct bitmap *);*/
-static long color_8888_bgr0(int);
-static long color_8888_0bgr(int);
-static long color_8888_0rgb(int);
-/*static void pass_0bgr(unsigned short *, struct bitmap *);*/
-static long color_555be(int);
-static long color_555(int);
-static long color_565be(int);
-static long color_565(int);
-/*static void make_8_table(int *, double);*/
-static void make_16_table(int *, int, int, float_double, int, int);
-static void make_red_table(int, int, int, int);
-static void make_green_table(int, int, int, int);
-static void make_blue_table(int, int, int, int);
-static void make_round_tables(void);
-
 int slow_fpu = -1;
 
-#define LIN \
-	r+=(int)(in[0]);\
-	g+=(int)(in[1]);\
-	b+=(int)(in[2]);\
-	in+=3;
-
 /* EMPIRE IMAGINE FEAR */
 #define LTABLES \
+	ir = in[0];\
+	ig = in[1];\
+	ib = in[2];\
+	r+=(int)ir;\
+	g+=(int)ig;\
+	b+=(int)ib;\
+	in+=3;\
 	{\
 		int rc=r,gc=g,bc=b;\
 		if ((unsigned)rc>65535) rc=rc<0?0:65535;\
 		if ((unsigned)gc>65535) gc=gc<0?0:65535;\
 		if ((unsigned)bc>65535) bc=bc<0?0:65535;\
-		if (table_16) {\
-			rt=red_table[rc];\
-			gt=green_table[gc];\
-			bt=blue_table[bc];\
-		} else {\
-			rt=red_table[rc >> 8];\
-			gt=green_table[gc >> 8];\
-			bt=blue_table[bc >> 8];\
-		}\
+		rt=red_table[rc >> shift];\
+		gt=green_table[gc >> shift];\
+		bt=blue_table[bc >> shift];\
 	}\
 	SAVE_CODE\
 	rt=r-(rt&65535);\
@@ -162,7 +128,6 @@ int slow_fpu = -1;
 
 
 #define BODY \
-	LIN\
 	LTABLES\
 	r=bptr[3];\
 	g=bptr[4];\
@@ -184,7 +149,6 @@ int slow_fpu = -1;
 	bptr[5]=bt;
 
 #define BODYR \
-	LIN\
 	LTABLES\
 	rt+=8;\
 	gt+=8;\
@@ -195,12 +159,11 @@ int slow_fpu = -1;
 	bptr[-3]+=3*rt;\
 	bptr[-2]+=3*gt;\
 	bptr[-1]+=3*bt;\
-	*bptr+=5*rt;\
+	bptr[0]+=5*rt;\
 	bptr[1]+=5*gt;\
 	bptr[2]+=5*bt;
 
 #define BODYC \
-	LIN\
 	LTABLES\
 	r=rt;\
 	g=gt;\
@@ -227,13 +190,15 @@ int slow_fpu = -1;
 	bptr[-1]+=3*bt;\
 	bptr+=3;
 
-#define DITHER_TEMPLATE(template_name) \
+#define DITHER_TEMPLATE(template_name, sh) \
 	static void template_name(unsigned short *my_restrict in, struct bitmap *out, int *dregs)\
-		{\
+	{\
+		const int shift = sh;\
+		unsigned short ir, ig, ib;\
 		int r,g,b,o,rt,gt,bt,y,x;\
 		unsigned char *my_restrict outp=out->data;\
 		int *my_restrict bptr;\
-		int skip=out->skip-SKIP_CODE;\
+		ssize_t skip = out->skip - SKIP_CODE;\
 \
 		o=0;o=o; /*warning go away */\
 		switch(out->x){\
@@ -262,36 +227,28 @@ int slow_fpu = -1;
 		}\
 	}
 
-#define ROUND_TEMPLATE(template_name)\
+#define ROUND_TEMPLATE(template_name, sh)\
 	static void template_name(unsigned short *my_restrict in, struct bitmap *out)\
 	{\
+		const int shift = sh;\
+		unsigned short ir, ig, ib;\
 		int rt,gt,bt,o,x,y;\
 		unsigned char *my_restrict outp=out->data;\
-		int skip=out->skip-SKIP_CODE;\
-	\
+		ssize_t skip = out->skip - SKIP_CODE;\
+\
 		o=0;o=o; /*warning go away */\
-		if (table_16) {\
-			for (y=out->y;y;y--){\
-				for (x=out->x;x;x--){\
-					rt=red_table[limit_16(in[0])];\
-					gt=green_table[limit_16(in[1])];\
-					bt=blue_table[limit_16(in[2])];\
-					in+=3;\
-					SAVE_CODE\
-				}\
-				outp+=skip;\
-			}\
-		} else {\
-			for (y=out->y;y;y--){\
-				for (x=out->x;x;x--){\
-					rt=red_table[limit_16(in[0]) >> 8];\
-					gt=green_table[limit_16(in[1]) >> 8];\
-					bt=blue_table[limit_16(in[2]) >> 8];\
-					in+=3;\
-					SAVE_CODE\
-				}\
-				outp+=skip;\
+		for (y=out->y;y;y--){\
+			for (x=out->x;x;x--){\
+				ir = in[0];\
+				ig = in[1];\
+				ib = in[2];\
+				rt=red_table[limit_16(ir) >> shift];\
+				gt=green_table[limit_16(ig) >> shift];\
+				bt=blue_table[limit_16(ib) >> shift];\
+				in+=3;\
+				SAVE_CODE\
 			}\
+			outp+=skip;\
 		}\
 	}
 
@@ -302,12 +259,40 @@ int slow_fpu = -1;
  * that saves appropriate code on *outp (unsigned char *outp). We can use int o;
  * as a scratchpad.
  */
-#define SAVE_CODE \
-	o=rt|gt|bt;\
-	*outp++=(o>>16);
+#define SAVE_CODE					\
+	o = (rt >> 16) + (gt >> 16) + (bt >> 16);	\
+	*outp++ = (unsigned char)o;
 
-DITHER_TEMPLATE(dither_1byte)
-ROUND_TEMPLATE(round_1byte)
+DITHER_TEMPLATE(dither_1byte, 0)
+ROUND_TEMPLATE(round_1byte, 0)
+DITHER_TEMPLATE(dither_1byte_8, 8)
+ROUND_TEMPLATE(round_1byte_8, 8)
+
+#undef SKIP_CODE
+#undef SAVE_CODE
+
+#define SKIP_CODE out->x
+#define SAVE_CODE						\
+	{							\
+		int rr, gr, br, or;				\
+		o = (rt >> 16) + (gt >> 16) + (bt >> 16);	\
+		rr = red_table[limit_16(ir) >> shift];		\
+		gr = green_table[limit_16(ig) >> shift];	\
+		br = blue_table[limit_16(ib) >> shift];		\
+		or = (rr >> 16) + (gr >> 16) + (br >> 16);	\
+		if (!((real_colors_table[or * 3 + 0] - ir) |	\
+		      (real_colors_table[or * 3 + 1] - ig) |	\
+		      (real_colors_table[or * 3 + 2] - ib))) {	\
+			o = or;					\
+		}						\
+		rt = real_colors_table[o * 3 + 0];		\
+		gt = real_colors_table[o * 3 + 1];		\
+		bt = real_colors_table[o * 3 + 2];		\
+		*outp++ = (unsigned char)o;			\
+	}							\
+
+DITHER_TEMPLATE(dither_1byte_real_colors, 0)
+DITHER_TEMPLATE(dither_1byte_real_colors_8, 8)
 
 #undef SKIP_CODE
 #undef SAVE_CODE
@@ -322,71 +307,88 @@ ROUND_TEMPLATE(round_1byte)
 #define SAVE_CODE \
 	o=rt|gt|bt;\
 	o>>=16;\
-	*(unsigned char *)outp=o;\
-	((unsigned char *)outp)[1]=o>>8;\
+	outp[0]=o;\
+	outp[1]=o>>8;\
 	outp+=2;
 #endif /* #ifdef t2c */
 
-DITHER_TEMPLATE(dither_2byte)
-ROUND_TEMPLATE(round_2byte)
+DITHER_TEMPLATE(dither_2byte, 0)
+ROUND_TEMPLATE(round_2byte, 0)
+DITHER_TEMPLATE(dither_2byte_8, 8)
+ROUND_TEMPLATE(round_2byte_8, 8)
 #undef SAVE_CODE
 #undef SKIP_CODE
 
 /* B G R */
-#define SKIP_CODE out->x*3;
-#define SAVE_CODE outp[0]=bt>>16;\
+#define SKIP_CODE out->x*3
+#define SAVE_CODE \
+	outp[0]=bt>>16;\
 	outp[1]=gt>>16;\
 	outp[2]=rt>>16;\
 	outp+=3;
-DITHER_TEMPLATE(dither_195)
-ROUND_TEMPLATE(round_195)
+DITHER_TEMPLATE(dither_195, 0)
+ROUND_TEMPLATE(round_195, 0)
+DITHER_TEMPLATE(dither_195_8, 8)
+ROUND_TEMPLATE(round_195_8, 8)
 #undef SAVE_CODE
 #undef SKIP_CODE
 
 /* R G B */
-#define SKIP_CODE out->x*3;
-#define SAVE_CODE *outp=rt>>16;\
+#define SKIP_CODE out->x*3
+#define SAVE_CODE \
+	outp[0]=rt>>16;\
 	outp[1]=gt>>16;\
 	outp[2]=bt>>16;\
 	outp+=3;
-DITHER_TEMPLATE(dither_451)
-ROUND_TEMPLATE(round_451)
+DITHER_TEMPLATE(dither_451, 0)
+ROUND_TEMPLATE(round_451, 0)
+DITHER_TEMPLATE(dither_451_8, 8)
+ROUND_TEMPLATE(round_451_8, 8)
 #undef SAVE_CODE
 #undef SKIP_CODE
 
 /* B G R 0 */
-#define SKIP_CODE out->x*4;
-#define SAVE_CODE *outp=bt>>16;\
+#define SKIP_CODE out->x*4
+#define SAVE_CODE \
+	outp[0]=bt>>16;\
 	outp[1]=gt>>16;\
 	outp[2]=rt>>16;\
 	outp[3]=0;\
 	outp+=4;
-DITHER_TEMPLATE(dither_196)
-ROUND_TEMPLATE(round_196)
+DITHER_TEMPLATE(dither_196, 0)
+ROUND_TEMPLATE(round_196, 0)
+DITHER_TEMPLATE(dither_196_8, 8)
+ROUND_TEMPLATE(round_196_8, 8)
 #undef SAVE_CODE
 #undef SKIP_CODE
 
 /* 0 B G R */
-#define SKIP_CODE out->x*4;
-#define SAVE_CODE *outp=0;\
+#define SKIP_CODE out->x*4
+#define SAVE_CODE \
+	outp[0]=0;\
 	outp[1]=bt>>16;\
 	outp[2]=gt>>16;\
 	outp[3]=rt>>16;\
 	outp+=4;
-DITHER_TEMPLATE(dither_452)
-ROUND_TEMPLATE(round_452)
+DITHER_TEMPLATE(dither_452, 0)
+ROUND_TEMPLATE(round_452, 0)
+DITHER_TEMPLATE(dither_452_8, 8)
+ROUND_TEMPLATE(round_452_8, 8)
 #undef SAVE_CODE
 #undef SKIP_CODE
 
 /* 0 R G B */
-#define SKIP_CODE out->x*4;
-#define SAVE_CODE *outp=0;\
+#define SKIP_CODE out->x*4
+#define SAVE_CODE \
+	outp[0]=0;\
 	outp[1]=rt>>16;\
 	outp[2]=gt>>16;\
 	outp[3]=bt>>16;\
 	outp+=4;
-DITHER_TEMPLATE(dither_708)
-ROUND_TEMPLATE(round_708)
+DITHER_TEMPLATE(dither_708, 0)
+ROUND_TEMPLATE(round_708, 0)
+DITHER_TEMPLATE(dither_708_8, 8)
+ROUND_TEMPLATE(round_708_8, 8)
 #undef SAVE_CODE
 #undef SKIP_CODE
 
@@ -410,23 +412,65 @@ static long color_332(int rgb)
 
 }
 
-static long color_121(int rgb)
+/* For 216-color cube */
+static long color_666(int rgb)
 {
-	int r,g,b;
+	int r, g, b;
+	unsigned char i;
 	long ret = 0;
 
-	r=(rgb>>16)&255;
-	g=(rgb>>8)&255;
-	b=rgb&255;
-	r=(r+127)/255;
-	g=(3*g+127)/255;
-	b=(b+127)/255;
-	*(unsigned char *)&ret=(r<<3)|(g<<1)|b;
+	r = (rgb >> 16) & 255;
+	g = (rgb >> 8) & 255;
+	b = rgb & 255;
+
+	r = (r * 5 + 127) / 255;
+	g = (g * 5 + 127) / 255;
+	b = (b * 5 + 127) / 255;
+
+	i = (unsigned char)r;
+	i *= 6;
+	i += g;
+	i *= 6;
+	i += b;
+
+	*(unsigned char *)&ret = i;
 	return ret;
 
 }
 
-static long color_pass_rgb(int rgb)
+static long color_121(int rgb)
+{
+	int r, g, b;
+	long ret = 0;
+
+	r = (rgb >> 16) & 255;
+	g = (rgb >> 8) & 255;
+	b = rgb & 255;
+	r = (r + 127) / 255;
+	g = (3 * g + 127) / 255;
+	b = (b + 127) / 255;
+	*(unsigned char *)&ret = (r << 3) | (g << 1) | b;
+	return ret;
+
+}
+
+static long color_111(int rgb)
+{
+	int r, g, b;
+	long ret = 0;
+
+	r = (rgb >> 16) & 255;
+	g = (rgb >> 8) & 255;
+	b = rgb & 255;
+	r = (r + 127) / 255;
+	g = (g + 127) / 255;
+	b = (b + 127) / 255;
+	*(unsigned char *)&ret = (r << 2) | (g << 1) | b;
+	return ret;
+
+}
+
+static long color_888_rgb(int rgb)
 {
 	long ret = 0;
 
@@ -449,26 +493,6 @@ static long color_888_bgr(int rgb)
 	return ret;
 }
 
-#if 0
-/* Long live the Manchester Modulation! */
-static void pass_bgr(unsigned short *in, struct bitmap *out)
-{
-	int skip=out->skip-3*out->x,y,x;
-	unsigned char *outp=out->data;
-
-	for (y=out->y;y;y--){
-		for (x=out->x;x;x--){
-			outp[0]=in[2];
-			outp[1]=in[1];
-			outp[2]=in[0];
-			outp+=3;
-			in+=3;
-		}
-		outp+=skip;
-	}
-}
-#endif
-
 static long color_8888_bgr0(int rgb)
 {
 	long ret = 0;
@@ -511,27 +535,6 @@ static long color_8888_0rgb(int rgb)
 	return ret;
 }
 
-#if 0
-/* We assume unsgned short holds at least 16 bits. */
-static void pass_0bgr(unsigned short *in, struct bitmap *out)
-{
-	int skip=out->skip-4*out->x,y,x;
-	unsigned char *outp=out->data;
-
-	for (y=out->y;y;y--){
-		for (x=out->x;x;x--){
-			outp[0]=0;
-			outp[1]=in[2]>>8;
-			outp[2]=in[1]>>8;
-			outp[3]=in[0]>>8;
-			outp+=4;
-			in+=3;
-		}
-		outp+=skip;
-	}
-}
-#endif
-
 /* We assume long holds at least 32 bits */
 static long color_555be(int rgb)
 {
@@ -608,6 +611,50 @@ static long color_565(int rgb)
 	return ret;
 }
 
+static long color_888_bgr_15bit(int rgb)
+{
+	int r,g,b;
+	long ret = 0;
+
+	r=(rgb>>16)&255;
+	g=(rgb>>8)&255;
+	/* Long live the PIN photodiode */
+	b=rgb&255;
+
+	r=(r*31+127)/255;
+	g=(g*31+127)/255;
+	b=(b*31+127)/255;
+
+	((unsigned char *)&ret)[0]=(unsigned char)(r<<3);
+	((unsigned char *)&ret)[1]=(unsigned char)(g<<3);
+	((unsigned char *)&ret)[2]=(unsigned char)(b<<3);
+
+	return ret;
+}
+
+static long color_888_bgr_16bit(int rgb)
+{
+	int r,g,b;
+	long ret = 0;
+
+	r=(rgb>>16)&255;
+	g=(rgb>>8)&255;
+	/* Long live the PIN photodiode */
+	b=rgb&255;
+
+	r=(r*31+127)/255;
+	g=(g*63+127)/255;
+	b=(b*31+127)/255;
+
+	((unsigned char *)&ret)[0]=(unsigned char)(r<<3);
+	((unsigned char *)&ret)[1]=(unsigned char)(g<<2);
+	((unsigned char *)&ret)[2]=(unsigned char)(b<<3);
+
+	return ret;
+}
+
+
+
 /* rgb = r*65536+g*256+b */
 /* The selected color_fn returns a long.
  * When we have for example 2 bytes per pixel, we make them in the memory,
@@ -616,16 +663,22 @@ static long color_565(int rgb)
  */
 long (*get_color_fn(int depth))(int rgb)
 {
-	switch(depth)
-	{
+	switch (depth) {
 		case 33:
 			return color_121;
 			break;
 
+		case 801:
+			return color_111;
+			break;
+
 		case 65:
 			return color_332;
 			break;
 
+		case 833:
+			return color_666;
+
 		case 122:
 			return color_555;
 			break;
@@ -643,7 +696,7 @@ long (*get_color_fn(int depth))(int rgb)
 			break;
 
 		case 451:
-			return color_pass_rgb;
+			return color_888_rgb;
 			break;
 
 		case 195:
@@ -662,6 +715,14 @@ long (*get_color_fn(int depth))(int rgb)
 			return color_8888_0rgb;
 			break;
 
+		case 15555:
+			return color_888_bgr_15bit;
+			break;
+
+		case 16579:
+			return color_888_bgr_16bit;
+			break;
+
 		default:
 			return NULL;
 			break;
@@ -669,44 +730,25 @@ long (*get_color_fn(int depth))(int rgb)
 	}
 }
 
-#if 0
-static void make_8_table(int *table, double gamma)
-{
-	int i,light0;
-	float_double light;
-	const float_double inv_255=1/255.;
-
-	for (i=0;i<256;i++){
-		light=fd_pow((float_double)i*inv_255,gamma);
-		/* Long live the Nipkow Disk */
-		light0=65535*light;
-		if (light0<0) light0=0;
-		if (light0>65535) light0=65535;
-		table[i]=light0;
-	}
-}
-#endif
-
 /* Gamma says that light=electricity raised to gamma */
-/* dump_t2c means memory organization defined in comment for
- * red_table on the top of dither.c */
-/* dump_t2c is taken into account only if t2c is defined. */
-static void make_16_table(int *table, int bits, int pos, float_double gamma, int dump_t2c,
-	int bigendian)
+static void make_16_table(int *table, int values, int mult, float_double gamma, int bigendian)
 {
-	int j,light_val,grades=(1<<bits)-1,grade;
+	int grades = values - 1;
+	int j, light_val, grade;
 	float_double voltage;
-	float_double rev_gamma=1/gamma;
-	const float_double inv_65535=(float_double)(1/65535.);
+	float_double rev_gamma = 1 / gamma;
+	const float_double inv_65535 = (float_double)(1 / 65535.);
 	int last_grade, last_content;
+	unsigned val;
 	uttime start_time = get_time();
 	int sample_state = 0;
 	int x_slow_fpu = slow_fpu;
+
 	if (gamma_bits != 2) x_slow_fpu = !gamma_bits;
 
-	repeat_loop:
-	last_grade=-1;
-	last_content=0;
+repeat_loop:
+	last_grade = -1;
+	last_content = 0;
 
 	for (j=0;j<65536;j++){
 		if (x_slow_fpu) {
@@ -729,18 +771,18 @@ static void make_16_table(int *table, int bits, int pos, float_double gamma, int
 				}
 			}
 		}
-		voltage=fd_pow(j*inv_65535,rev_gamma);
+		voltage=fd_pow(j * inv_65535, rev_gamma);
 		/* Determine which monitor input voltage is equivalent
 		 * to said photon flux level
 		 */
 
-		grade=(int)(voltage*grades+(float_double)0.5);
-		if (grade==last_grade){
-			table[j]=last_content;
+		grade=(int)(voltage * grades + (float_double)0.5);
+		if (grade == last_grade){
+			table[j] = last_content;
 			continue;
 		}
-		last_grade=grade;
-		voltage=(float_double)grade/grades;
+		last_grade = grade;
+		voltage = (float_double)grade / grades;
 		/* Find nearest voltage to this voltage. Finding nearest voltage, not
 		 * nearest photon flux ensures the dithered pixels will be perceived to be
 		 * near. The voltage input into the monitor was intentionally chosen by
@@ -750,23 +792,19 @@ static void make_16_table(int *table, int bits, int pos, float_double gamma, int
 		 * kool ;-) (and is kool)
 		 */
 
-		light_val=(int)(fd_pow(voltage,gamma)*65535+(float_double)0.5);
+		light_val = (int)(fd_pow(voltage, gamma) * 65535 + (float_double)0.5);
 		/* Find out what photon flux this index represents */
 
-		if (light_val<0) light_val=0;
-		if (light_val>65535) light_val=65535;
+		if (light_val < 0) light_val = 0;
+		if (light_val > 65535) light_val = 65535;
 		/* Clip photon flux for safety */
 
-		if (bigendian) {
-			unsigned val, val2;
-			val = grade<<pos;
-			val2 = (val>>8) | ((val&0xff)<<8);
-			last_content=light_val|((unsigned)val2<<16U);
-		}else{
-			last_content=light_val|((unsigned)grade<<(pos+16U));
-		}
+		val = grade * mult;
+		if (bigendian)
+			val = (val >> 8) | ((val & 0xff) << 8);
+		last_content = light_val | (val << 16);
 
-		table[j]=last_content;
+		table[j] = last_content;
 		/* Save index and photon flux. */
 	}
 	if (x_slow_fpu == -1) slow_fpu = 0;	/* if loop passed once without
@@ -774,22 +812,22 @@ static void make_16_table(int *table, int bits, int pos, float_double gamma, int
 	if (gamma_bits == 2 && x_slow_fpu == 1) slow_fpu = 1;
 }
 
-static void make_red_table(int bits, int pos, int dump_t2c, int be)
+static void make_red_table(int values, int mult, int be)
 {
 	red_table = mem_realloc(red_table, 65536 * sizeof(*red_table));
-	make_16_table(red_table,bits,pos,(float_double)display_red_gamma,dump_t2c, be);
+	make_16_table(red_table, values, mult, (float_double)(display_red_gamma tcc_hack), be);
 }
 
-static void make_green_table(int bits, int pos, int dump_t2c, int be)
+static void make_green_table(int values, int mult, int be)
 {
 	green_table = mem_realloc(green_table, 65536 * sizeof(*green_table));
-	make_16_table(green_table,bits,pos,(float_double)display_green_gamma,dump_t2c, be);
+	make_16_table(green_table, values, mult, (float_double)(display_green_gamma tcc_hack), be);
 }
 
-static void make_blue_table(int bits, int pos,int dump_t2c, int be)
+static void make_blue_table(int values, int mult, int be)
 {
 	blue_table = mem_realloc(blue_table, 65536 * sizeof(*blue_table));
-	make_16_table(blue_table,bits,pos,(float_double)display_blue_gamma, dump_t2c, be);
+	make_16_table(blue_table, values, mult, (float_double)(display_blue_gamma tcc_hack), be);
 }
 
 void dither(unsigned short *in, struct bitmap *out)
@@ -825,12 +863,12 @@ static void make_round_tables(void)
 	/* ICC bug */
 	icc_volatile unsigned short v;
 
-	for (a=0;a<256;a++){
+	for (a = 0; a < 256; a++) {
 		/* a is sRGB coordinate */
-		v=ags_8_to_16((unsigned char)a,(float)((float)user_gamma/(float)sRGB_gamma));
-		round_red_table[a]=red_table[v >> (8 - 8 * table_16)] & 0xffff;
-		round_green_table[a]=green_table[v >> (8 - 8 * table_16)] & 0xffff;
-		round_blue_table[a]=blue_table[v >> (8 - 8 * table_16)] & 0xffff;
+		v=ags_8_to_16((unsigned char)a, (float)(user_gamma / sRGB_gamma));
+		round_red_table[a] = red_table[v >> (8 - 8 * table_16)] & 0xffff;
+		round_green_table[a] = green_table[v >> (8 - 8 * table_16)] & 0xffff;
+		round_blue_table[a] = blue_table[v >> (8 - 8 * table_16)] & 0xffff;
 	}
 }
 
@@ -865,65 +903,107 @@ static void compress_tables(void)
 	blue_table = bt;
 }
 
+static void make_real_colors_table(void)
+{
+	unsigned short *real_colors;
+	if (real_colors_table) mem_free(real_colors_table), real_colors_table = NULL;
+	if (round_fn != round_1byte && round_fn != round_1byte_8)
+		return;
+	if (!drv->get_real_colors)
+		return;
+	real_colors = drv->get_real_colors();
+	if (!real_colors)
+		return;
+	real_colors_table = mem_alloc(256 * 3 * sizeof(unsigned short));
+	agx_48_to_48(real_colors_table, real_colors, 256, (float)(display_red_gamma tcc_hack), (float)(display_green_gamma tcc_hack), (float)(display_blue_gamma tcc_hack));
+	mem_free(real_colors);
+}
+
 /* Also makes up the dithering tables.
  * You may call it twice - it doesn't leak any memory.
  */
 void init_dither(int depth)
 {
 	table_16 = 1;
-	switch(depth){
+	switch (depth) {
 		case 33:
 		/* 4bpp, 1Bpp */
-		make_red_table(1,3,0,0);
-		make_green_table(2,1,0,0);
-		make_blue_table(1,0,0,0);
-		dither_fn_internal=dither_1byte;
-		round_fn=round_1byte;
+		make_red_table(1 << 1, 1 << 3, 0);
+		make_green_table(1 << 2, 1 << 1, 0);
+		make_blue_table(1 << 1, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_1byte : dither_1byte_8;
+		round_fn = table_16 ? round_1byte : round_1byte_8;
+		break;
+
+		case 801:
+		/* 8bpp, 1Bpp, 1x1x1 */
+		make_red_table(1 << 1, 1 << 2, 0);
+		make_green_table(1 << 1, 1 << 1, 0);
+		make_blue_table(1 << 1, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_1byte : dither_1byte_8;
+		round_fn = table_16 ? round_1byte : round_1byte_8;
 		break;
 
 		case 65:
-		/* 8 bpp, 1 Bpp */
-		make_red_table(3,5,0,0);
-		make_green_table(3,2,0,0);
-		make_blue_table(2,0,0,0);
-		dither_fn_internal=dither_1byte;
-		round_fn=round_1byte;
+		/* 8 bpp, 1Bpp */
+		make_red_table(1 << 3, 1 << 5, 0);
+		make_green_table(1 << 3, 1 << 2, 0);
+		make_blue_table(1 << 2, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_1byte : dither_1byte_8;
+		round_fn = table_16 ? round_1byte : round_1byte_8;
+		break;
+
+		case 833:
+		/* 8bpp, 1Bpp, 6x6x6 */
+		make_red_table(6, 36, 0);
+		make_green_table(6, 6, 0);
+		make_blue_table(6, 1, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_1byte : dither_1byte_8;
+		round_fn = table_16 ? round_1byte : round_1byte_8;
 		break;
 
 		case 122:
 		/* 15bpp, 2Bpp */
-		make_red_table(5,10,1,0);
-		make_green_table(5,5,1,0);
-		make_blue_table(5,0,1,0);
-		dither_fn_internal=dither_2byte;
-		round_fn=round_2byte;
+		make_red_table(1 << 5, 1 << 10, 0);
+		make_green_table(1 << 5, 1 << 5, 0);
+		make_blue_table(1 << 5, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_2byte : dither_2byte_8;
+		round_fn = table_16 ? round_2byte : round_2byte_8;
 		break;
 
 		case 378:
-		/* 15bpp, 2Bpp, disordered (I have a mental disorder) */
-		make_red_table(5,10,1,1);
-		make_green_table(5,5,1,1);
-		make_blue_table(5,0,1,1);
-		dither_fn_internal=dither_2byte;
-		round_fn=round_2byte;
+		/* 15bpp, 2Bpp, disordered (1 << I have a mental disorder) */
+		make_red_table(1 << 5, 1 << 10, 1);
+		make_green_table(1 << 5, 1 << 5, 1);
+		make_blue_table(1 << 5, 1 << 0, 1);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_2byte : dither_2byte_8;
+		round_fn = table_16 ? round_2byte : round_2byte_8;
 		break;
 
 		case 130:
 		/* 16bpp, 2Bpp */
-		make_red_table(5,11,1,0);
-		make_green_table(6,5,1,0);
-		make_blue_table(5,0,1,0);
-		dither_fn_internal=dither_2byte;
-		round_fn=round_2byte;
+		make_red_table(1 << 5, 1 << 11, 0);
+		make_green_table(1 << 6, 1 << 5, 0);
+		make_blue_table(1 << 5, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_2byte : dither_2byte_8;
+		round_fn = table_16 ? round_2byte : round_2byte_8;
 		break;
 
 		case 386:
 		/* 16bpp, 2Bpp, disordered */
-		make_red_table(5,11,1,1);
-		make_green_table(6,5,1,1);
-		make_blue_table(5,0,1,1);
-		dither_fn_internal=dither_2byte;
-		round_fn=round_2byte;
+		make_red_table(1 << 5, 1 << 11, 1);
+		make_green_table(1 << 6, 1 << 5, 1);
+		make_blue_table(1 << 5, 1 << 0, 1);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_2byte : dither_2byte_8;
+		round_fn = table_16 ? round_2byte : round_2byte_8;
 		break;
 
 		case 451:
@@ -931,11 +1011,12 @@ void init_dither(int depth)
 		 * Even this is dithered!
 		 * R G B
 		 */
-		make_red_table(8,0,0,0);
-		make_green_table(8,0,0,0);
-		make_blue_table(8,0,0,0);
-		dither_fn_internal=dither_451;
-		round_fn=round_451;
+		make_red_table(1 << 8, 1 << 0, 0);
+		make_green_table(1 << 8, 1 << 0, 0);
+		make_blue_table(1 << 8, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_451 : dither_451_8;
+		round_fn = table_16 ? round_451 : round_451_8;
 		break;
 
 		case 195:
@@ -943,11 +1024,12 @@ void init_dither(int depth)
 		 * Even this is dithered!
 		 * B G R
 		 */
-		make_red_table(8,0,0,0);
-		make_green_table(8,0,0,0);
-		make_blue_table(8,0,0,0);
-		dither_fn_internal=dither_195;
-		round_fn=round_195;
+		make_red_table(1 << 8, 1 << 0, 0);
+		make_green_table(1 << 8, 1 << 0, 0);
+		make_blue_table(1 << 8, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_195 : dither_195_8;
+		round_fn = table_16 ? round_195 : round_195_8;
 		break;
 
 		case 452:
@@ -955,11 +1037,12 @@ void init_dither(int depth)
 		 * Even this is dithered!
 		 * 0 B G R
 		 */
-		make_red_table(8,0,0,0);
-		make_green_table(8,0,0,0);
-		make_blue_table(8,0,0,0);
-		dither_fn_internal=dither_452;
-		round_fn=round_452;
+		make_red_table(1 << 8, 1 << 0, 0);
+		make_green_table(1 << 8, 1 << 0, 0);
+		make_blue_table(1 << 8, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_452 : dither_452_8;
+		round_fn = table_16 ? round_452 : round_452_8;
 		break;
 
 		case 196:
@@ -967,11 +1050,12 @@ void init_dither(int depth)
 		 * Even this is dithered!
 		 * B G R 0
 		 */
-		make_red_table(8,0,0,0);
-		make_green_table(8,0,0,0);
-		make_blue_table(8,0,0,0);
-		dither_fn_internal=dither_196;
-		round_fn=round_196;
+		make_red_table(1 << 8, 1 << 0, 0);
+		make_green_table(1 << 8, 1 << 0, 0);
+		make_blue_table(1 << 8, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_196 : dither_196_8;
+		round_fn = table_16 ? round_196 : round_196_8;
 		break;
 
 		case 708:
@@ -979,19 +1063,55 @@ void init_dither(int depth)
 		 * Even this is dithered!
 		 * 0 R G B
 		 */
-		make_red_table(8,0,0,0);
-		make_green_table(8,0,0,0);
-		make_blue_table(8,0,0,0);
-		dither_fn_internal=dither_708;
-		round_fn=round_708;
+		make_red_table(1 << 8, 1 << 0, 0);
+		make_green_table(1 << 8, 1 << 0, 0);
+		make_blue_table(1 << 8, 1 << 0, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_708 : dither_708_8;
+		round_fn = table_16 ? round_708 : round_708_8;
+		break;
+
+		case 15555:
+		/* 24bpp, 3Bpp, downsampled to 15bit
+		 * B G R
+		 */
+		make_red_table(1 << 5, 1 << 3, 0);
+		make_green_table(1 << 5, 1 << 3, 0);
+		make_blue_table(1 << 5, 1 << 3, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_195 : dither_195_8;
+		round_fn = table_16 ? round_195 : round_195_8;
+		break;
+
+		case 16579:
+		/* 24bpp, 3Bpp, downsampled to 16bit
+		 * B G R
+		 */
+		make_red_table(1 << 5, 1 << 3, 0);
+		make_green_table(1 << 6, 1 << 2, 0);
+		make_blue_table(1 << 5, 1 << 3, 0);
+		compress_tables();
+		dither_fn_internal = table_16 ? dither_195 : dither_195_8;
+		round_fn = table_16 ? round_195 : round_195_8;
 		break;
 
 		default:
-		internal("Graphics driver returned unsupported \
-pixel memory organisation %d",depth);
+		internal_error("Graphics driver returned unsupported pixel memory organisation %d",depth);
 	}
-	compress_tables();
+
 	make_round_tables();
+	make_real_colors_table();
+
+	if (real_colors_table) {
+		if (dither_fn_internal == dither_1byte)
+			dither_fn_internal = dither_1byte_real_colors;
+		if (dither_fn_internal == dither_1byte_8)
+			dither_fn_internal = dither_1byte_real_colors_8;
+	}
+
+	gamma_cache_rgb_1 = -2;
+	gamma_cache_rgb_2 = -2;
+	gamma_stamp++;
 }
 
 /* Input is in sRGB space (unrounded, i. e. directly from HTML)
@@ -1001,9 +1121,20 @@ pixel memory organisation %d",depth);
 void round_color_sRGB_to_48(unsigned short *my_restrict red, unsigned short *my_restrict green,
 		unsigned short *my_restrict blue, int rgb)
 {
-	*red=round_red_table[(rgb>>16)&255];
-	*green=round_green_table[(rgb>>8)&255];
-	*blue=round_blue_table[rgb&255];
+	*red = round_red_table[(rgb >> 16) & 255];
+	*green = round_green_table[(rgb >> 8) & 255];
+	*blue = round_blue_table[rgb & 255];
+	if (real_colors_table) {
+		int shift, rt, gt, bt, o;
+		shift = 8 - table_16 * 8;
+		rt = red_table[*red >> shift];
+		gt = green_table[*green >> shift];
+		bt = blue_table[*blue >> shift];
+		o = (rt >> 16) + (gt >> 16) + (bt >> 16);
+		*red = real_colors_table[o * 3 + 0];
+		*green = real_colors_table[o * 3 + 1];
+		*blue = real_colors_table[o * 3 + 2];
+	}
 }
 
 void free_dither(void)
@@ -1011,6 +1142,7 @@ void free_dither(void)
 	if (red_table) mem_free(red_table), red_table = DUMMY;
 	if (green_table) mem_free(green_table), green_table = DUMMY;
 	if (blue_table) mem_free(blue_table), blue_table = DUMMY;
+	if (real_colors_table) mem_free(real_colors_table), real_colors_table = NULL;
 }
 
 #endif
