@@ -71,11 +71,11 @@ static void get_c_width(unsigned char *attr, int *w, int sh)
 	unsigned char *al;
 	if ((al = get_attr_val(attr, cast_uchar "width"))) {
 		if (*al && al[strlen(cast_const_char al) - 1] == '*') {
-			unsigned char *en;
+			char *end;
 			unsigned long n;
 			al[strlen(cast_const_char al) - 1] = 0;
-			n = strtoul(cast_const_char al, (char **)(void *)&en, 10);
-			if (n < 10000 && !*en) *w = W_REL - (int)n;
+			n = strtoul(cast_const_char al, &end, 10);
+			if (n < 10000 && !*end) *w = W_REL - (int)n;
 		} else {
 			int p = get_width(attr, cast_uchar "width", sh);
 			if (p >= 0) *w = p;
@@ -110,7 +110,6 @@ struct table_cell {
 #ifdef G
 	unsigned char bgcolor_str[8];
 	struct g_object_area *root;
-	struct g_object_tag *tag_object;
 	struct rect_set *brd;
 	int g_width;
 	struct rect rect;
@@ -149,6 +148,8 @@ struct table {
 	int rh;
 	int link_num;
 	struct rgb bgcolor;
+	struct list_head **row_tags;
+	int row_tags_n;
 #ifdef G
 	struct background *bg;
 	struct background *frame_bg;
@@ -159,12 +160,14 @@ struct table {
 	struct table_cell ***r_cells;
 	int *w_cells;
 	int nr_cells;
+	struct g_object_tag **tags;
+	int tags_n;
 #endif
 };
 
 
 #ifdef DEBUG
-#define CELL(t, x, y) (((x) < 0 || (x) >= (t)->rx || (y) < 0 || (y) >= (t)->ry) ? (internal("accessing cell out of table (%d,%d) - limit (%d,%d)", (x), (y), (t)->rx, (t)->ry), (t)->cells) : &(t)->cells[(y) * (t)->rx + (x)])
+#define CELL(t, x, y) (((x) < 0 || (x) >= (t)->rx || (y) < 0 || (y) >= (t)->ry) ? (internal_error("accessing cell out of table (%d,%d) - limit (%d,%d)", (x), (y), (t)->rx, (t)->ry), (t)->cells) : &(t)->cells[(y) * (t)->rx + (x)])
 #else
 #define CELL(t, x, y) (&(t)->cells[(y) * (t)->rx + (x)])
 #endif
@@ -200,6 +203,8 @@ static struct table *new_table(void)
 	t->r_cells = DUMMY;
 	t->w_cells = DUMMY;
 	t->nr_cells = 0;
+	t->tags = DUMMY;
+	t->tags_n = 0;
 #endif
 	t->x = t->y = 0;
 	t->rx = INIT_X;
@@ -211,6 +216,8 @@ static struct table *new_table(void)
 	t->xcols = DUMMY;
 	t->xc = 0;
 	t->r_heights = DUMMY;
+	t->row_tags = DUMMY;
+	t->row_tags_n = 0;
 	return t;
 }
 
@@ -222,7 +229,6 @@ static void free_table(struct table *t)
 		if (c->tag) mem_free(c->tag);
 #ifdef G
 		if (c->root) c->root->go.destruct(&c->root->go);
-		if (c->tag_object) c->tag_object->go.destruct(&c->tag_object->go);
 		if (c->brd) mem_free(c->brd);
 #endif
 	}
@@ -235,6 +241,9 @@ static void free_table(struct table *t)
 		for (i = 0; i < t->nr_cells; i++) mem_free(t->r_cells[i]);
 		mem_free(t->r_cells), mem_free(t->w_cells);
 		if (t->frame_bg) g_release_background(t->frame_bg);
+		for (i = 0; i < t->tags_n; i++)
+			t->tags[i]->go.destruct(&t->tags[i]->go);
+		mem_free(t->tags);
 	}
 #endif
 	if (t->bordercolor) mem_free(t->bordercolor);
@@ -245,6 +254,11 @@ static void free_table(struct table *t)
 	mem_free(t->cols);
 	mem_free(t->xcols);
 	mem_free(t->cells);
+	for (i = 0; i < t->row_tags_n; i++) if (t->row_tags[i]) {
+		free_list(struct tag, *t->row_tags[i]);
+		mem_free(t->row_tags[i]);
+	}
+	mem_free(t->row_tags);
 	mem_free(t);
 }
 
@@ -377,7 +391,7 @@ unsigned char *skip_element(unsigned char *html, unsigned char *eof, unsigned ch
 	int namelen;
 	r:
 	while (html < eof && (*html != '<')) rr:html++;
-	if (html + 2 <= eof && (html[1] == '!' || html[1] == '?')) {
+	if (eof - html >= 2 && (html[1] == '!' || html[1] == '?')) {
 		html = skip_comment(html, eof);
 		goto r;
 	}
@@ -388,6 +402,34 @@ unsigned char *skip_element(unsigned char *html, unsigned char *eof, unsigned ch
 	goto r;
 }
 
+static void get_row_tag(struct table *t, int y, unsigned char *t_attr)
+{
+	unsigned char *a;
+	struct tag *tag;
+	a = get_attr_val(t_attr, cast_uchar "id");
+	if (!a)
+		return;
+	while (y >= t->row_tags_n) {
+		int i;
+		int new_n = t->row_tags_n * 2;
+		if (!new_n)
+			new_n = 1;
+		if ((unsigned)new_n > MAXINT / sizeof(struct list_head *)) overalloc();
+		t->row_tags = mem_realloc(t->row_tags, new_n * sizeof(struct list_head *));
+		for (i = t->row_tags_n; i < new_n; i++)
+			t->row_tags[i] = NULL;
+		t->row_tags_n = new_n;
+	}
+	if (!t->row_tags[y]) {
+		t->row_tags[y] = mem_alloc(sizeof(struct list_head));
+		init_list(*t->row_tags[y]);
+	}
+	tag = mem_alloc(sizeof(struct tag) + strlen(cast_const_char a));
+	strcpy(cast_char tag->name, cast_const_char a);
+	mem_free(a);
+	add_to_list_end(*t->row_tags[y], tag);
+}
+
 struct s_e {
 	unsigned char *s, *e;
 };
@@ -414,6 +456,7 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 	int i, j, k;
 	struct rgb l_col;
 	int c_al = AL_TR, c_val = VAL_TR, c_width = W_AUTO, c_span = 0;
+	int norow = 1;
 	memcpy(&l_col, bgcolor, sizeof(struct rgb));
 	*end = html;
 	if (bad_html) {
@@ -427,7 +470,7 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 	see:
 	html = en;
 	if (bad_html && !p && !lbhp) {
-		if (!(*bhp & (ALLOC_GR-1))) {
+		if (!(*bhp & (ALLOC_GR - 1))) {
 			if ((unsigned)*bhp > MAXINT / sizeof(struct s_e) - ALLOC_GR) overalloc();
 			*bad_html = mem_realloc(*bad_html, (*bhp + ALLOC_GR) * sizeof(struct s_e));
 		}
@@ -439,7 +482,7 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 		if (lbhp) (*bad_html)[*bhp-1].e = html;
 		goto scan_done;
 	}
-	if (html + 2 <= eof && (html[1] == '!' || html[1] == '?')) {
+	if (eof - html >= 2 && (html[1] == '!' || html[1] == '?')) {
 		html = skip_comment(html, eof);
 		goto se;
 	}
@@ -464,6 +507,7 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 		goto scan_done;
 	}
 	if (t_namelen == 8 && !casecmp(t_name, cast_uchar "COLGROUP", 8)) {
+		get_row_tag(t, y + 1, t_attr);
 		if (c_span) new_columns(t, c_span, c_width, c_al, c_val, 1);
 		if (lbhp) (*bad_html)[*bhp-1].e = html, lbhp = NULL;
 		c_al = AL_TR;
@@ -486,6 +530,7 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 	}
 	if (t_namelen == 3 && !casecmp(t_name, cast_uchar "COL", 3)) {
 		int sp, wi, al, val;
+		get_row_tag(t, y + 1, t_attr);
 		if (lbhp) (*bad_html)[*bhp-1].e = html, lbhp = NULL;
 		if ((sp = get_num(t_attr, cast_uchar "span")) == -1) sp = 1;
 		wi = c_width;
@@ -498,12 +543,14 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 		c_span = 0;
 		goto see;
 	}
+	if (t_namelen == 3 && !casecmp(t_name, cast_uchar "/TR", 3)) norow = 1;
 	if (t_namelen == 3 && (!casecmp(t_name, cast_uchar "/TR", 3) || !casecmp(t_name, cast_uchar "/TD", 3) || !casecmp(t_name, cast_uchar "/TH", 3))) {
 		if (c_span) new_columns(t, c_span, c_width, c_al, c_val, 1);
 		if (p) CELL(t, x, y)->end = html, p = 0;
 		if (lbhp) (*bad_html)[*bhp-1].e = html, lbhp = NULL;
 	}
 	if (t_namelen == 2 && !casecmp(t_name, cast_uchar "TR", 2)) {
+		get_row_tag(t, y + 1, t_attr);
 		if (c_span) new_columns(t, c_span, c_width, c_al, c_val, 1);
 		if (p) CELL(t, x, y)->end = html, p = 0;
 		if (lbhp) (*bad_html)[*bhp-1].e = html, lbhp = NULL;
@@ -515,9 +562,11 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 		get_valign(t_attr, &l_val);
 		get_bgcolor(t_attr, &l_col);
 		y++, x = 0;
+		norow = 0;
 		goto see;
 	}
 	if (t_namelen == 5 && ((!casecmp(t_name, cast_uchar "THEAD", 5)) || (!casecmp(t_name, cast_uchar "TBODY", 5)) || (!casecmp(t_name, cast_uchar "TFOOT", 5)))) {
+		get_row_tag(t, y + 1, t_attr);
 		if (c_span) new_columns(t, c_span, c_width, c_al, c_val, 1);
 		if (lbhp) (*bad_html)[*bhp-1].e = html, lbhp = NULL;
 		group = 2;
@@ -527,8 +576,9 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 	if (lbhp) (*bad_html)[*bhp-1].e = html, lbhp = NULL;
 	if (p) CELL(t, x, y)->end = html, p = 0;
 	if (y == -1) y = 0, x = 0;
+	if (norow && x) y++, x = 0, norow = 0;
 	nc:
-	if (!(cell = new_cell(t, x, y))) goto see;
+	cell = new_cell(t, x, y);
 	if (cell->used) {
 		if (cell->colspan == -1) goto see;
 		x++;
@@ -594,7 +644,8 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 	qqq = t->x;
 	for (i = 1; csp != -1 ? i < csp : x + i < qqq; i++) {
 		struct table_cell *sc;
-		if (!(sc = new_cell(t, x + i, y)) || sc->used) {
+		sc = new_cell(t, x + i, y);
+		if (sc->used) {
 			csp = i;
 			for (k = 0; k < i; k++) CELL(t, x + k, y)->colspan = csp;
 			break;
@@ -609,10 +660,11 @@ static struct table *parse_table(unsigned char *html, unsigned char *eof, unsign
 	for (j = 1; rsp != -1 ? j < rsp : y + j < qqq; j++) {
 		for (k = 0; k < i; k++) {
 			struct table_cell *sc;
-			if (!(sc = new_cell(t, x + k, y + j)) || sc->used) {
+			sc = new_cell(t, x + k, y + j);
+			if (sc->used) {
 				int l, m;
 				if (sc->mx == x && sc->my == y) continue;
-				/*internal("boo");*/
+				/*internal_error("boo");*/
 				for (l = 0; l < k; l++) memset(CELL(t, x + l, y + j), 0, sizeof(struct table_cell));
 				rsp = j;
 				for (l = 0; l < i; l++) for (m = 0; m < j; m++) CELL(t, x + l, y + m)->rowspan = j;
@@ -671,7 +723,7 @@ static void get_cell_width(struct table *t, struct table_cell *c, int w, int a,
 #endif
 	}
 	/*debug("get_cell_width: %d < %d", *min, *max);*/
-	/*if (min && max && *min > *max) internal("get_cell_width: %d > %d", *min, *max);*/
+	/*if (min && max && *min > *max) internal_error("get_cell_width: %d > %d", *min, *max);*/
 }
 
 #define g_c_w(cc)							\
@@ -708,7 +760,7 @@ static void dst_width(int *p, int n, int w, int *lim)
 		if (lim && p[i] > lim[i]) w = safe_add(w, p[i] - lim[i]), p[i] = lim[i];
 	}
 	if (w) {
-		/*if (!lim) internal("bug in dst_width");*/
+		/*if (!lim) internal_error("bug in dst_width");*/
 		lim = NULL;
 		s = 0;
 		goto again;
@@ -824,7 +876,7 @@ static int get_column_widths(struct table *t)
 			struct table_cell *c = CELL(t, i, j);
 			if (c->spanned || !c->used) continue;
 			if (c->colspan + i > t->x) {
-				/*internal("colspan out of table");
+				/*internal_error("colspan out of table");
 				return -1;*/
 				continue;
 			}
@@ -882,7 +934,7 @@ static void get_table_width(struct table *t)
 	}
 	t->min_t = min;
 	t->max_t = max;
-	/*if (min > max) internal("min(%d) > max(%d)", min, max);*/
+	/*if (min > max) internal_error("min(%d) > max(%d)", min, max);*/
 }
 
 static void distribute_widths(struct table *t, int width)
@@ -896,7 +948,7 @@ static void distribute_widths(struct table *t, int width)
 	t->rw = 0;
 	if (!t->x) return;
 	if (d < 0) {
-		/*internal("too small width %d, required %d", width, t->min_t);*/
+		/*internal_error("too small width %d, required %d", width, t->min_t);*/
 		return;
 	}
 	for (i = 0; i < t->x; i++) if (t->max_c[i] > mmax_c) mmax_c = t->max_c[i];
@@ -949,7 +1001,7 @@ static void distribute_widths(struct table *t, int width)
 					w[i] = 1, mx[i] = MAXINT;
 					break;
 				default:
-					/*internal("could not expand table");*/
+					/*internal_error("could not expand table");*/
 					goto end2;
 			}
 			p = safe_add(p, w[i]);
@@ -977,7 +1029,7 @@ static void distribute_widths(struct table *t, int width)
 			d -= t->w_c[mii] - q;
 			while (d < 0) t->w_c[mii]--, d++;
 			if (t->w_c[mii] < q) {
-				/*internal("shrinking cell");*/
+				/*internal_error("shrinking cell");*/
 				t->w_c[mii] = q;
 			}
 			wq = 1;
@@ -1009,7 +1061,7 @@ static void check_table_widths(struct table *t)
 		if (c->x_width > p) {
 			/*int min, max;
 			get_cell_width(t, c, 0, 0, &min, &max, NULL);
-			internal("cell is now wider (%d > %d) min = %d, max = %d, now_min = %d, now_max = %d", c->x_width, p, t->min_c[i], t->max_c[i], min, max);*/
+			internal_error("cell is now wider (%d > %d) min = %d, max = %d, now_min = %d, now_max = %d", c->x_width, p, t->min_c[i], t->max_c[i], min, max);*/
 			/* sbohem, internale. chytl jsi mi spoustu chyb v tabulkovaci, ale ted je proste cas jit ... ;-( */
 			c->x_width = p;
 		}
@@ -1021,7 +1073,7 @@ static void check_table_widths(struct table *t)
 			struct table_cell *c = CELL(t, i, j);
 			if (!c->start) continue;
 			if (c->colspan + i > t->x) {
-				/*internal("colspan out of table");*/
+				/*internal_error("colspan out of table");*/
 				mem_free(w);
 				return;
 			}
@@ -1036,7 +1088,7 @@ static void check_table_widths(struct table *t)
 					c = 0;
 					for (l = i; l < i + s; l++) if (w[l] < t->w_c[k]) w[l] = safe_add(w[l], 1), w[k]--, c = 1;
 					if (w[k] > t->w_c[k]) {
-						if (!c) internal("can't shrink cell");
+						if (!c) internal_error("can't shrink cell");
 						else goto ag;
 					}
 				}*/
@@ -1054,7 +1106,7 @@ static void check_table_widths(struct table *t)
 		}*/
 	}
 	if (ns > s) {
-		/*internal("new width(%d) is larger than previous(%d)", ns, s);*/
+		/*internal_error("new width(%d) is larger than previous(%d)", ns, s);*/
 		mem_free(w);
 		return;
 	}
@@ -1114,7 +1166,7 @@ static void get_table_heights(struct table *t)
 #ifdef G
 			} else {
 				if (!(gp = g_format_html_part(cell->start, cell->end, cell->align, 0, xw, NULL, cell->link_num, NULL /* FIX: background image */, cell->bgcolor_str, t->gp->data))) return;
-				/*if (gp->root->xw > xw) internal("vono to neumi formatovat... buuuu (%d > %d)", gp->root->xw, xw);*/
+				/*if (gp->root->xw > xw) internal_error("vono to neumi formatovat... buuuu (%d > %d)", gp->root->xw, xw);*/
 				cell->root = gp->root;
 				gp->root = NULL;
 				cell->height = gp->y;
@@ -1153,10 +1205,27 @@ static void get_table_heights(struct table *t)
 		}
 	} while ((s = ns) != MAXINT);
 	if (!F) {
-		t->rh = (!!(t->frame & F_ABOVE) + !!(t->frame & F_BELOW)) * !!t->border;
+		t->rh = !!(t->frame & F_ABOVE) * !!t->border;
 		for (j = 0; j < t->y; j++) {
-			t->rh = safe_add(t->rh, t->r_heights[j]);
 			if (j && get_hline_width(t, j) >= 0) t->rh = safe_add(t->rh, 1);
+			if (j < t->row_tags_n && t->row_tags[j]) {
+				struct tag *tag;
+				struct list_head *ltag;
+				foreach(struct tag, tag, ltag, *t->row_tags[j]) {
+					tag->x = 0;
+					tag->y = t->rh;
+				}
+			}
+			t->rh = safe_add(t->rh, t->r_heights[j]);
+		}
+		t->rh = safe_add(t->rh, !!(t->frame & F_BELOW) * !!t->border);
+		if (j < t->row_tags_n && t->row_tags[j]) {
+			struct tag *tag;
+			struct list_head *ltag;
+			foreach(struct tag, tag, ltag, *t->row_tags[j]) {
+				tag->x = 0;
+				tag->y = t->rh;
+			}
 		}
 #ifdef G
 	} else {
@@ -1164,6 +1233,14 @@ static void get_table_heights(struct table *t)
 		for (j = 0; j <= t->y; j++) {
 			int qw = g_get_hline_pad(t, j, NULL, NULL);
 			t->rh = safe_add(t->rh, qw);
+			if (j < t->row_tags_n && t->row_tags[j]) {
+				struct tag *tag;
+				struct list_head *ltag;
+				foreach(struct tag, tag, ltag, *t->row_tags[j]) {
+					tag->x = 0;
+					tag->y = t->rh;
+				}
+			}
 			if (j < t->y) t->rh = safe_add(t->rh, t->r_heights[j]);
 		}
 #endif
@@ -1175,6 +1252,17 @@ static void display_complicated_table(struct table *t, int x, int y, int *yy)
 	int i, j;
 	struct f_data *f = t->p->data;
 	int yp, xp = safe_add(x, (t->frame & F_LHS) && t->border);
+	if (f) {
+		for (j = 0; j <= t->y; j++) {
+			if (j < t->row_tags_n && t->row_tags[j]) {
+				struct tag *tag;
+				struct list_head *ltag;
+				foreach(struct tag, tag, ltag, *t->row_tags[j]) {
+					html_tag(f, tag->name, safe_add(x, tag->x), safe_add(y, tag->y));
+				}
+			}
+		}
+	}
 	for (i = 0; i < t->x; i++) {
 		yp = safe_add(y, (t->frame & F_ABOVE) && t->border);
 		for (j = 0; j < t->y; j++) {
@@ -1561,7 +1649,7 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	nn->y = safe_add(p->yp, cye);
 	nn->xw = n->xw;
 	add_to_list(p->data->nodes, nn);
-	/*if (p->cy + t->rh != cye) internal("size does not match; 1:%d, 2:%d", p->cy + t->rh, cye);*/
+	/*if (p->cy + t->rh != cye) internal_error("size does not match; 1:%d, 2:%d", p->cy + t->rh, cye);*/
 	p->cy = cye;
 
 	ret2:
@@ -1601,7 +1689,7 @@ static void add_to_cell_sets(struct table_cell ****s, int **nn, int *n, struct r
 {
 	int i, j;
 	if (r->y1 < 0 || r->y2 < 0)
-		internal("add_to_cell_sets: integer overflow: %d, %d", r->y1, r->y2);
+		internal_error("add_to_cell_sets: integer overflow: %d, %d", r->y1, r->y2);
 	for (i = r->y1 >> RECT_BOUND_BITS; i <= (r->y2 - 1) >> RECT_BOUND_BITS; i++) {
 		if (i >= *n) {
 			struct table_cell ***ns;
@@ -1706,11 +1794,11 @@ static void table_get_list(struct g_object *o_, void (*fn)(struct g_object *pare
 	int i, j;
 	for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) {
 		struct table_cell *c = CELL(t, i, j);
-		if (c->tag_object)
-			fn(&o->go, &c->tag_object->go);
 		if (c->root)
 			fn(&o->go, &c->root->go);
 	}
+	for (i = 0; i < t->tags_n; i++)
+		fn(&o->go, &t->tags[i]->go);
 }
 
 void table_bg(struct text_attrib *ta, unsigned char bgstr[8])
@@ -1725,6 +1813,26 @@ void table_bg(struct text_attrib *ta, unsigned char bgstr[8])
 	} else sprintf(cast_char bgstr, "#%02x%02x%02x", G_HTML_TABLE_FRAME_COLOR, G_HTML_TABLE_FRAME_COLOR, G_HTML_TABLE_FRAME_COLOR);
 }
 
+static void g_new_tag(struct table *t, unsigned char *name, int x, int y)
+{
+	struct g_object_tag *tag;
+	size_t sl;
+	sl = strlen(cast_const_char name);
+	if (sl > MAXINT - sizeof(struct g_object_tag)) overalloc();
+	tag = mem_calloc(sizeof(struct g_object_tag) + sl);
+	tag->go.mouse_event = g_dummy_mouse;
+	tag->go.draw = g_dummy_draw;
+	tag->go.destruct = g_tag_destruct;
+	strcpy(cast_char tag->name, cast_const_char name);
+	tag->go.x = x;
+	tag->go.y = y;
+	if (!(t->tags_n & (ALLOC_GR - 1))) {
+		if ((unsigned)t->tags_n > MAXINT / sizeof(struct g_object_tag *) - ALLOC_GR) overalloc();
+		t->tags = mem_realloc(t->tags, (t->tags_n + ALLOC_GR) * sizeof(struct g_object_tag *));
+	}
+	t->tags[t->tags_n++] = tag;
+}
+
 static void process_g_table(struct g_part *gp, struct table *t)
 {
 	int i, j;
@@ -1889,23 +1997,19 @@ static void process_g_table(struct g_part *gp, struct table *t)
 	gp->cx = -1;
 	gp->cx_w = 0;
 
-	for (j = 0; j < t->y; j++) {
-		for (i = 0; i < t->x; i++) {
+	for (j = 0; j <= t->y; j++) {
+		if (j < t->row_tags_n && t->row_tags[j]) {
+			struct tag *tag;
+			struct list_head *ltag;
+			foreach(struct tag, tag, ltag, *t->row_tags[j]) {
+				g_new_tag(t, tag->name, safe_add(o->go.xw, tag->x), safe_add(o->go.yw, tag->y));
+			}
+		}
+		if (j < t->y) for (i = 0; i < t->x; i++) {
 			struct table_cell *c;
 			c = CELL(t, i, j);
 			if (c->root && c->tag) {
-				struct g_object_tag *tag;
-				size_t sl;
-				sl = strlen(cast_const_char c->tag);
-				if (sl > MAXINT - sizeof(struct g_object_tag)) overalloc();
-				tag = mem_calloc(sizeof(struct g_object_tag) + sl);
-				tag->go.mouse_event = g_dummy_mouse;
-				tag->go.draw = g_dummy_draw;
-				tag->go.destruct = g_tag_destruct;
-				strcpy(cast_char tag->name, cast_const_char c->tag);
-				tag->go.x = c->root->go.x;
-				tag->go.y = c->root->go.y + o->go.yw;
-				c->tag_object = tag;
+				g_new_tag(t, c->tag, safe_add(o->go.xw, c->root->go.x), safe_add(o->go.yw, c->root->go.y));
 			}
 		}
 	}
