@@ -31,6 +31,7 @@ static_const struct {
 		{"tn3270", 0, NULL, tn3270_func,	0, 0, 0, 0, 1},
 		{"mms", 0, NULL, mms_func,		1, 0, 1, 0, 1},
 		{"magnet", 0, NULL, magnet_func,	1, 0, 0, 0, 1},
+		{"gopher", 0, NULL, gopher_func,	1, 0, 0, 0, 1},
 #ifdef JS
 		{"javascript", 0, NULL, javascript_func,1, 0, 0, 0, 0},
 #endif
@@ -266,6 +267,7 @@ static void translate_directories(unsigned char *url)
 	int lo = !casecmp(url, cast_uchar "file://", 7);
 	if (!casecmp(url, cast_uchar "javascript:", 11)) return;
 	if (!casecmp(url, cast_uchar "magnet:", 7)) return;
+	if (!casecmp(url, cast_uchar "gopher:", 7)) return;
 	if (!dd || dd == url /*|| *--dd != '/'*/) return;
 	if (!dsep(*dd)) {
 		dd--;
@@ -352,7 +354,7 @@ static unsigned char *translate_hashbang(unsigned char *up)
 static unsigned char *rewrite_url_google_docs(unsigned char *n)
 {
 	int i;
-	unsigned char *id, *id_end;
+	unsigned char *id, *id_end, *url_end;
 	unsigned char *res;
 	int l;
 	struct {
@@ -361,9 +363,13 @@ static unsigned char *rewrite_url_google_docs(unsigned char *n)
 		const char *result2;
 	} const patterns[] = {
 		{ "https://docs.google.com/document/d/", "https://docs.google.com/document/d/", "/export?format=pdf" },
+		{ "https://docs.google.com/document/u/", "https://docs.google.com/document/u/", "/export?format=pdf" },
 		{ "https://docs.google.com/spreadsheets/d/", "https://docs.google.com/spreadsheets/d/", "/export?format=pdf" },
+		{ "https://docs.google.com/spreadsheets/u/", "https://docs.google.com/spreadsheets/u/", "/export?format=pdf" },
 		{ "https://docs.google.com/presentation/d/", "https://docs.google.com/presentation/d/", "/export/pdf" },
-		{ "https://drive.google.com/file/d/", "https://drive.google.com/uc?export=download&id=", "" }
+		{ "https://docs.google.com/presentation/u/", "https://docs.google.com/presentation/u/", "/export/pdf" },
+		{ "https://drive.google.com/file/d/", "https://drive.google.com/uc?export=download&id=", "" },
+		{ "https://drive.google.com/file/u/", "https://drive.google.com/uc?export=download&id=", "" }
 	};
 	for (i = 0; i < (int)array_elements(patterns); i++) {
 		if (!cmpbeg(n, cast_uchar patterns[i].beginning))
@@ -372,11 +378,17 @@ static unsigned char *rewrite_url_google_docs(unsigned char *n)
 	return n;
 match:
 	id = n + strlen(cast_const_char patterns[i].beginning);
-	id_end = cast_uchar strchr(cast_const_char id, '/');
+	url_end = id + strcspn(cast_const_char id, "#" POST_CHAR_STRING);
+	id_end = memrchr(id, '/', url_end - id);
 	if (!id_end)
 		return n;
 	if (!cmpbeg(id_end, cast_uchar "/export"))
 		return n;
+	if (!patterns[i].result2[0]) {
+		id = id_end;
+		while (id[-1] != '/')
+			id--;
+	}
 	res = init_str();
 	l = 0;
 	add_to_str(&res, &l, cast_uchar patterns[i].result1);
@@ -413,15 +425,71 @@ static unsigned char *rewrite_url(unsigned char *n)
 	return n;
 }
 
+static int test_qualified_name(unsigned char *host, unsigned char *hostname)
+{
+	unsigned char *c;
+	if (!casestrcmp(host, hostname))
+		return 1;
+	c = cast_uchar strchr(cast_const_char hostname, '.');
+	if (c) {
+		*c = 0;
+		if (!casestrcmp(host, hostname))
+			return 1;
+	}
+	return 0;
+}
+
+static int is_local_host(unsigned char *host)
+{
+	if (!*host)
+		return 1;
+	if (!casestrcmp(host, cast_uchar "localhost"))
+		return 1;
+#if defined(HAVE_GETHOSTNAME)
+	{
+		int rs;
+		unsigned char n[4096];
+		n[0] = 0;
+		EINTRLOOP(rs, gethostname(cast_char n, sizeof n));
+		n[sizeof n - 1] = 0;
+		if (!rs && strlen(cast_const_char n) < sizeof n - 1) {
+			if (test_qualified_name(host, n))
+				return 1;
+		}
+	}
+#elif defined(HAVE_SYS_UTSNAME_H) && defined(HAVE_UNAME)
+	{
+		int rs;
+		struct utsname name;
+		memset(&name, 0, sizeof name);
+		EINTRLOOP(rs, uname(&name));
+		if (rs >= 0) {
+			if (test_qualified_name(host, cast_uchar name.nodename))
+				return 1;
+		}
+	}
+#endif
+	return 0;
+}
+
 static void insert_wd(unsigned char **up, unsigned char *cwd)
 {
 	unsigned char *u = *up;
 	unsigned char *cw;
 	unsigned char *url;
+	unsigned char *host;
 	int url_l;
+	int i;
 	if (!u || !cwd || !*cwd) return;
 	if (casecmp(u, cast_uchar "file://", 7)) return;
-	if (dir_sep(u[7])) return;
+	for (i = 7; u[i] && !dir_sep(u[i]); i++) ;
+	host = memacpy(u + 7, i - 7);
+	if (is_local_host(host)) {
+		mem_free(host);
+		memmove(u + 7, u + i, strlen(cast_const_char (u + i)) + 1);
+		return;
+	}
+	mem_free(host);
 #ifdef DOS_FS
 	if (upcase(u[7]) >= 'A' && upcase(u[7]) <= 'Z' && u[8] == ':' && dir_sep(u[9])) return;
 #endif
@@ -510,7 +578,7 @@ unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 		if (!(s = cast_uchar strstr(cast_const_char base, "//"))) {
 			if (!(s = cast_uchar strchr(cast_const_char base, ':'))) {
 				bad_base:
-				internal("bad base url: %s", base);
+				internal_error("bad base url: %s", base);
 				return NULL;
 			}
 			s++;
@@ -1143,6 +1211,8 @@ static unsigned char *display_url_or_host(struct terminal *term, unsigned char *
 		url_enc = idn_encode_url(url_conv2, 0);
 	else
 		url_enc = idn_encode_host(url_conv2, (int)strlen(cast_const_char url_conv2), separator, 0);
+	if (!url_enc)
+		url_enc = stracpy(url_conv2), is_idn = 1;
 	mem_free(url_conv2);
 	if (!strcmp(cast_const_char url_enc, cast_const_char url)) {
 		if (is_idn && warn_idn) {
